use lazy_static::lazy_static;
use crate::config::MAX_APP_NUM;
use crate::sync::UPSafeCell;
use crate::task::task::{TaskControlBlock, TaskStatus};
use crate::loader::{get_num_app, init_app_cx};
use crate::println;
use crate::task::context::TaskContext;

mod context;
mod switch;
mod task;


// 公开到外部的一个全局任务管理器的结构体
pub struct TaskManager {
    num_app: usize,  // app的数量 这个在os运行之后就不会有变化
    inner: UPSafeCell<TaskManagerInner>,  // 这个内部TaskControlBlock 会随着系统运行发生变化
}

impl TaskManager {
    fn run_first_task(&self){
        println!("run_first_task");
    }
}


// 不公开的结构体, 有一个MAX_APP_NUM大小的数组, 用来保存TCB, 和当前任务的的TCB的下标
struct TaskManagerInner {
    tasks: [TaskControlBlock; MAX_APP_NUM],
    current_task: usize,
}

lazy_static!{
    pub static ref TASK_MANAGER: TaskManager = {
        // 得到app的总数
        let num_app = get_num_app();

        // 初始化内核中的任务列表
        let mut tasks = [
            TaskControlBlock {
                task_cx: TaskContext::new(),
                // UnInit 这个tcb 还没用户应用填充加载
                task_status: TaskStatus::UnInit
            };
            MAX_APP_NUM
        ];

        // 初始化用户的应用对应的 TCB
        for i in 0..num_app {
            // 初始化用户应用的内核栈
            let kernel_app_stack_ptr = init_app_cx(i);

            // 从 初始内核栈的栈顶(保存 trap_context的位置) 创建 任务context
            let task_context = TaskContext::from(kernel_app_stack_ptr);

            // 把内核栈的信息保存到 TCB中
            tasks[i].task_cx = task_context;

            // 初始为 准备好的状态
            tasks[i].task_status = TaskStatus::Ready
        };

        TaskManager {
            num_app,
            inner: unsafe {
                UPSafeCell::new(TaskManagerInner {
                    tasks,
                    current_task: 0,
                })
            }
        }
    };
}


pub fn run_first_task() {
    TASK_MANAGER.run_first_task();
}
